//-----------------------------------------------------------------------------
//
// ndx.cpp [ gyabo 2009 (http://www012.upp.so-net.ne.jp/gyabo/) ]
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// Include
//-----------------------------------------------------------------------------
#include <stdio.h>
#include <windows.h>
#include <mmsystem.h>
#include <d3d9.h>
#include <d3dx9.h>

//-----------------------------------------------------------------------------
// Pragma (VC only)
//-----------------------------------------------------------------------------
#pragma comment(lib, "winmm.lib")
#pragma comment(lib, "user32.lib")
#pragma comment(lib, "d3d9.lib")
#pragma comment(lib, "d3dx9.lib")

//-----------------------------------------------------------------------------
// MACRO
//-----------------------------------------------------------------------------
#define RELEASE(x)         { if(x) (x)->Release(); x = NULL; }
#define D3DCLEAR_DEFAULT   (D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER)

//-----------------------------------------------------------------------------
// Enumerate
//-----------------------------------------------------------------------------
enum {
  Width  = 1280,
  Height = 1024,
};

//-----------------------------------------------------------------------------
// data kz
//-----------------------------------------------------------------------------
struct shader {
  struct {
    LPD3DXBUFFER            buffer;
    LPDIRECT3DVERTEXSHADER9 shader;
  } vs;
  
  struct {
    LPD3DXBUFFER            buffer;
    LPDIRECT3DPIXELSHADER9  shader;
  } ps;
};

//-----------------------------------------------------------------------------
// Variable
//-----------------------------------------------------------------------------
static D3DPRESENT_PARAMETERS d3dpp = {
  Width, Height, D3DFMT_A8R8G8B8, 2, D3DMULTISAMPLE_NONE, 0,
  D3DSWAPEFFECT_DISCARD, 0, FALSE, TRUE, D3DFMT_D24X8,  0, 0, TRUE,
};

static float               fTime     = 0.0f;
static HWND                hWnd      = NULL;
static LPDIRECT3D9         d3d9      = NULL;
static LPDIRECT3DDEVICE9   d3ddevice = NULL;

//Resource
static LPD3DXMESH    mesh      = NULL;
static char          vs_tech[] = "vs_0";
static char          ps_tech[] = "ps_0";
static char       *shader_file = "shader.fx";

static shader        shader00;

void InitShader() {
  HRESULT      hRet = E_FAIL;
  D3DXCompileShaderFromFile(
    shader_file, 0, 0, vs_tech,
    D3DXGetVertexShaderProfile(d3ddevice),
    0, &shader00.vs.buffer, 0, 0);
  
  D3DXCompileShaderFromFile(
    shader_file, 0, 0, ps_tech,
    D3DXGetPixelShaderProfile(d3ddevice),
    0, &shader00.ps.buffer, 0, 0);
  
  hRet = d3ddevice->CreateVertexShader(
    (const DWORD*)shader00.vs.buffer->GetBufferPointer(),
    &shader00.vs.shader
  );
  if( FAILED(hRet) ) {
    printf("CreateVertexShader:Failed");
  }

  hRet = d3ddevice->CreatePixelShader(
    (const DWORD*)shader00.ps.buffer->GetBufferPointer(),
    &shader00.ps.shader
  );
  if( FAILED(hRet) ) {
    printf("CreatePixelShader:Failed");
  }
}

void TermShader() {
  RELEASE(shader00.vs.buffer);
  RELEASE(shader00.vs.shader);
  RELEASE(shader00.ps.buffer);
  RELEASE(shader00.ps.shader);
}

int InitRender() {
  //Create Box
  LPD3DXMESH temp = NULL;
  LPD3DXMESH temp2 = NULL;
  D3DXCreateBox(d3ddevice, 3, 3, 3, &temp, 0);
  temp->CloneMeshFVF(D3DXMESH_32BIT, D3DFVF_XYZ | D3DFVF_NORMAL, d3ddevice, &temp2);
  D3DXTessellateNPatches(temp2, NULL, 4.0f, FALSE, &mesh, NULL);
  D3DXComputeNormals(mesh, NULL);
  RELEASE(temp2);
  RELEASE(temp);
  
  //Create Shader
  InitShader();
  
  return 0;
}

void TermRender() {
  TermShader();
  RELEASE(mesh);
}


void OnLost() {
  if(shader00.vs.shader) {
    RELEASE(shader00.ps.shader)
  }
  
  if(shader00.ps.shader) {
    RELEASE(shader00.ps.shader)
  }
}

void OnRestore() {
  InitShader();
}

static D3DXMATRIX mView;
static D3DXMATRIX mProj;

void DrawMesh(D3DXVECTOR3 &vPos, D3DXVECTOR3 &vScale, D3DXVECTOR3 &vRot, D3DXVECTOR4 &vColor) {
  D3DXMATRIX mWorldViewProj;
  D3DXMATRIX mWorld;
  D3DXMATRIX mTrans;
  D3DXMATRIX mScale;
  D3DXMATRIX mRot;
  
  D3DXMatrixTranslation(&mTrans, vPos.x, vPos.y, vPos.z);
  D3DXMatrixScaling(&mScale, vScale.x, vScale.y, vScale.z);
  D3DXMatrixRotationYawPitchRoll(&mRot, vRot.x, vRot.y, vRot.z);
  mWorld = mScale * mRot * mTrans;
  
  mWorldViewProj = mWorld * mView * mProj;
  D3DXMatrixTranspose(&mWorldViewProj, &mWorldViewProj);
  d3ddevice->SetVertexShaderConstantF(16, (FLOAT*)&mWorldViewProj, 4);
  
  //Setup Material
  d3ddevice->SetVertexShaderConstantF(2, (FLOAT*)&vColor, 1);
  
  //Draw
  mesh->DrawSubset(0);
}

//-----------------------------------------------------------------------------
// Do Render
//-----------------------------------------------------------------------------
void DoRender() {
  while(GetAsyncKeyState(VK_ESCAPE) == 0) {
    
    fTime += 0.01677777f;
    
    if(d3ddevice->TestCooperativeLevel() == D3DERR_DEVICENOTRESET) {
      OnLost();
      d3ddevice->Reset(&d3dpp);
      OnRestore();
    }
    
    if(d3ddevice->BeginScene() == D3D_OK) {
      d3ddevice->Clear(NULL, 0, D3DCLEAR_DEFAULT, 0x00775566, 1.0f, 0);
      //To do render
      static FLOAT angle = 0;
      angle += 0.01f;
      FLOAT xpos = cos(angle * 0.8f) * 10;
      FLOAT ypos = -sin(angle * 0.7f) * 10;
      FLOAT zpos = sin(angle * 0.9f) * 10;
      FLOAT fAspect = (FLOAT)Width / (FLOAT)Height;

      
      //D3DXVECTOR3 vPos(0.0f, 0.0f, -50.0f);
      D3DXVECTOR3 vPos(xpos, ypos, zpos);
      D3DXVECTOR3 vAt(0.0f, 0.0f,  0.0f);
      D3DXVECTOR3 vUp(0.0f, 1.0f,  0.0f);
      D3DXMatrixLookAtLH(&mView, &vPos, &vAt, &vUp);
      D3DXMatrixPerspectiveFovLH(&mProj, D3DX_PI / 3.0f, fAspect, 1.0f / 32.0f, 256.0f);
      
      //Setup Light
      D3DXVECTOR4 vlight(0.6f, -1.0f, 1.6f, 0.0f);
      d3ddevice->SetVertexShaderConstantF(0, (FLOAT*)&vlight, 1);
      
      //Setup EyePos
      d3ddevice->SetVertexShaderConstantF(1, (FLOAT*)&vPos, 1);
      
      //Setup Time
      d3ddevice->SetVertexShaderConstantF(32, &fTime, 1);
      
      //Set Shader
      d3ddevice->SetVertexShader(shader00.vs.shader);
      d3ddevice->SetPixelShader(shader00.ps.shader);
      
      int sq = 2;
      for(int k = -sq; k <= sq; k++) {
        for(int j = -sq; j <= sq; j++) {
          for(int i = -sq; i <= sq; i++) {
            DrawMesh(
              D3DXVECTOR3(i * 20, j * 20, k * 20), 
              D3DXVECTOR3(1.0f, 1.0f, 1.0f), 
              D3DXVECTOR3(i, j + fTime, k), 
              D3DXVECTOR4(1.0f, 1.0f, 1.0f, 1.0f));
          }
        }
      }
      d3ddevice->EndScene();
      d3ddevice->Present(0, 0, 0, 0);
    }
    Sleep(0); //need...
  }
}



//-----------------------------------------------------------------------------
// Main (Entry point)
//-----------------------------------------------------------------------------
int main() {
  //Get Edit Control
  hWnd = CreateWindowExA(0, "edit", 0, 0, 0, 0, Width, Height, 0, 0, 0, 0);
  ShowWindow(hWnd, TRUE);
  
  //Create DirectGraphics
  d3d9 = Direct3DCreate9(D3D_SDK_VERSION);
  
  if(hWnd && d3d9) {
    ShowCursor(FALSE);
    d3d9->CreateDevice(
      D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
      D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3dpp, &d3ddevice);
    if(d3ddevice) {
      InitRender();
      //Start Render
      DoRender();
    }
  }
  ShowCursor(TRUE);
  
  TermRender();
  
  //Terminate
  RELEASE(d3ddevice);
  RELEASE(d3d9);
  return 0;
}
